#ifndef OGSE_HBAO_H
#define OGSE_HBAO_H

#define g_Resolution 	float2(1280.f, 1024.f)
#define g_InvResolution float2(1/1280.f, 1/1024.f)
#ifndef SSAO_QUALITY
	#define	SSAO_QUALITY int(1)
#endif
static const float g_R = HBAO_RADIUS;
static const float g_sqr_R = g_R*g_R;
static const float g_inv_R = 1/g_R;

static const float g_NumDir = 3 + SSAO_QUALITY;

#define M_PI 3.14159265f

//----------------------------------------------------------------------------------
float tangent(float3 P, float3 S)
{
    return (P.z - S.z) / length(S.xy - P.xy);
}

float3 tangent_eye_pos(float2 uv, float4 tangentPlane)
{
    // view vector going through the surface point at uv
    float3 V = tex2D(s_position, uv);
    float NdotV = dot(tangentPlane.xyz, V);
    // intersect with tangent plane except for silhouette edges
    if (NdotV < 0.0) V *= (tangentPlane.w / NdotV);
    return V;
}

float length2(float3 v) { return dot(v, v); } 

//----------------------------------------------------------------------------------
float3 min_diff(float3 P, float3 Pr, float3 Pl)
{
    float3 V1 = Pr - P;
    float3 V2 = P - Pl;
    return (length2(V1) < length2(V2)) ? V1 : V2;
}

//----------------------------------------------------------------------------------
float falloff(float r)
{
    return 1.0f - r*r;
}

//----------------------------------------------------------------------------------
float2 snap_uv_offset(float2 uv)
{
    return round(uv * g_Resolution) * g_InvResolution;
}

//----------------------------------------------------------------------------------
float horizon_occlusion(float2 deltaUV, 
                        float2 uv0, 
                        float3 P, 
                        float numSteps, 
                        float randstep,
                        float3 dPdu,
                        float3 dPdv)
{
    // Randomize starting point within the first sample distance
    float2 uv = uv0 + snap_uv_offset( randstep * deltaUV );
    
    // Snap increments to pixels to avoid disparities between xy 
    // and z sample locations and sample along a line
    deltaUV = snap_uv_offset( deltaUV );

    // Compute tangent vector using the tangent plane
    float3 T = deltaUV.x * dPdu + deltaUV.y * dPdv;

    float phi = atan(-T.z / length(T.xy)) + HBAO_ANGLE_BIAS;
    float tanH = tan(min(phi, M_PI*0.5));
    float sinH = tanH / sqrt(1.0f + tanH*tanH);

    float ao = 0;
    for(float j = 1; j <= numSteps; ++j) {
    uv += deltaUV;
        float3 S = tex2D(s_position, uv);
        
        // Ignore any samples outside the radius of influence
        float d2  = length2(S - P);
        if (d2 < g_sqr_R) {
            float tanS = tangent(P, S);

            if(tanS > tanH) {
                // Accumulate AO between the horizon and the sample
                float sinS = tanS / sqrt(1.0f + tanS*tanS);
                float r = sqrt(d2) * g_inv_R;
                ao += falloff(r) * (sinS - sinH)*saturate((abs(P.z - S.z) - HBAO_THRESHOLD)*1000);
                
                // Update the current horizon angle
                tanH = tanS;
                sinH = sinS;
            }
        }
    }

    return ao;
}

float3 calc_hbao(float3 P, float3 N, float2 tc0)
{
    // Calculate the real number of steps based on Z distance, and  
    // early out if geometry is too far away.   	

	float2 	step_size	= float2(.5f / 1280.h, .5f / 1024.h)*200/max(P.z, HBAO_STEP_SIZE);
    float numDirs = min ( g_NumDir, min(step_size.x / 4 * g_Resolution.x, step_size.y / 4 * g_Resolution.y));
    step_size = step_size / ( HBAO_NUM_STEPS + 1 );

    float4 tangentPlane = float4(N, dot(P, N));
    
    // Nearest neighbor pixels on the tangent plane
    float3 Pr = tangent_eye_pos(tc0 + float2(g_InvResolution.x, 0), tangentPlane);
    float3 Pl = tangent_eye_pos(tc0 + float2(-g_InvResolution.x, 0), tangentPlane);
    float3 Pt = tangent_eye_pos(tc0 + float2(0, g_InvResolution.y), tangentPlane);
    float3 Pb = tangent_eye_pos(tc0 + float2(0, -g_InvResolution.y), tangentPlane);

    // Screen-aligned basis for the tangent plane
    float3 dPdu = min_diff(P, Pr, Pl);
    float3 dPdv = min_diff(P, Pt, Pb) * (g_Resolution.y * g_InvResolution.x);

    // Loop over all directions
    float ao = 0;
    float delta = 0.6;//g_NumDir / numDirs;
    float alpha = 2.0f * M_PI / g_NumDir;
    for (float d = 0; d < g_NumDir; d+=delta) {
        float angle = alpha * d;
        float2 dir = float2(cos(angle), sin(angle));
        float2 deltaUV = dir * step_size.xy;
        ao += horizon_occlusion(deltaUV, tc0, P, HBAO_NUM_STEPS, 1, dPdu, dPdv);
    }

	ao /= g_NumDir;
	ao = saturate(1 - ao);
	return (ao + (1 - ao)*(1 - HBAO_BLEND_FACTOR));
}
#endif